/*

	BoxPainter

	Author:	Vendicator
	
	Description:
	Class that encapsulates box skinning of a window.
	Either uses a bitmap or colors (border/middle bg).

	Requires:

	Changelog:
	[2003-07-19 - Vendicator]
	- First work, split out from systray code (by Maduin/grd).
	
*/

#include "BoxPainter.h"

#include "../current/lsapi/lsapi.h"

#define MAGICPINK RGB(255, 0, 255)

BoxPainter::BoxPainter() :
	hbmSkin(NULL),
	hbmBack(NULL),
	hrgnBack(NULL),
	hPaintWnd(NULL),
	bTransBack(false),
	bSkinTiled(false),
	nWidth(0),
	nHeight(0)
{
}

BoxPainter::~BoxPainter()
{
	CleanUp();
}

bool BoxPainter::IsTransparent() const
{
	return bTransBack;
}

void BoxPainter::SetWindow(HWND hwnd)
{
	hPaintWnd = hwnd;
}

void BoxPainter::SetSize(int width, int height)
{
	nWidth = width;
	nHeight = height;
}

void BoxPainter::SetTiled(bool tile)
{
	bSkinTiled = tile;
}

void BoxPainter::SetBorders(int nLeft, int nTop, int nRight, int nBottom)
{
	nBorderLeft = nLeft;
	nBorderTop = nTop;
	nBorderRight = nRight;
	nBorderBottom = nBottom;
}

void BoxPainter::CleanUp()
{
	// delete gdi objects...
	bBitmap = false;

	if (hbmSkin)
	{
		DeleteObject( hbmSkin );
		hbmSkin = NULL;
	}
	if (hbmBack)
	{
		DeleteObject( hbmBack );
		hbmBack = NULL;
	}
	if (hrgnBack)
	{
		DeleteObject( hrgnBack );
		hrgnBack = NULL;
	}
}

void BoxPainter::SetColors(COLORREF clrBg, COLORREF clrBrdr)
{
	bBitmap = false;
	clrBack = clrBg;
	clrBorder = clrBrdr;

	bTransSkin = (clrBack == MAGICPINK);
	if (!bTransSkin)
		bTransSkin = (nBorderTop + nBorderLeft + nBorderRight + nBorderBottom > 0) && (clrBorder == MAGICPINK);
}

// handle to the bitmap used for painting
void BoxPainter::SetBitmap(HBITMAP hbmBg)
{
	BITMAP bm;
	HRGN hBitmapRgn, hOpaqueRgn;

	CleanUp();

	bBitmap = true;
	hbmSkin = hbmBg;

	// get info from the bitmap
	GetObject( hbmSkin, sizeof(BITMAP), &bm );

	hBitmapRgn = BitmapToRegion( hbmSkin, RGB(255, 0, 255), 0x101010, 0, 0);
	hOpaqueRgn = CreateRectRgn( 0, 0, bm.bmWidth, bm.bmHeight );

	// to see if any pixels are transparent
	bTransSkin = !EqualRgn( hBitmapRgn, hOpaqueRgn );

	DeleteObject( hBitmapRgn );
	DeleteObject( hOpaqueRgn );
}

bool BoxPainter::CreateBackground()
{
	HDC hdcScreen, hdcDst;
	HBITMAP hbmDstOld;

	// delete old objects if necessary...
	if ( hbmBack )
		DeleteObject( hbmBack );
	hbmBack = NULL;

	if ( hrgnBack )
		DeleteObject( hrgnBack );
	hrgnBack = NULL;

#ifdef _DEBUG
	if (hPaintWnd == NULL)
		MessageBox(NULL, TEXT("BoxPainter could not create a background cause it does not know which window it belongs to"), TEXT("BoxPainter Initialization Error!"), MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
#endif

	// initiate a new destination dc
	hdcScreen = GetDC(hPaintWnd);

	hdcDst = CreateCompatibleDC(hdcScreen);
	hbmBack = CreateCompatibleBitmap(hdcScreen, nWidth, nHeight);
	hbmDstOld = (HBITMAP)SelectObject(hdcDst, hbmBack);

	ReleaseDC(hPaintWnd, hdcScreen);

	// if the background colors are needed put them on...
	if ((!hbmSkin) || bTransSkin)
	{
		RECT r;
		HBRUSH brush;

		// fill the border with the border colors
		if (nBorderTop || nBorderLeft || nBorderBottom || nBorderTop)
		{
			brush = CreateSolidBrush(clrBorder);

			if (nBorderTop)
			{
				SetRect(&r, 0, 0, nWidth, nBorderTop);
				FillRect(hdcDst, &r, brush);
			}
			if (nBorderLeft)
			{
				SetRect(&r, 0, nBorderTop, nBorderLeft, nHeight - nBorderBottom);
				FillRect(hdcDst, &r, brush);
			}
			if (nBorderRight)
			{
				SetRect(&r, nWidth - nBorderRight, nBorderTop, nWidth, nHeight - nBorderBottom);
				FillRect(hdcDst, &r, brush);
			}
			if (nBorderBottom)
			{
				SetRect(&r, 0, nHeight - nBorderBottom, nWidth, nHeight);
				FillRect(hdcDst, &r, brush);
			}
			DeleteObject(brush);
		}

		// now fill the inner rectangle (excluding the borders) with the background color
		brush = CreateSolidBrush(clrBack);

		SetRect(&r, nBorderLeft, nBorderTop, nWidth - nBorderRight, nHeight - nBorderBottom);
		FillRect(hdcDst, &r, brush);

		DeleteObject(brush);
	}

	// if a background image is provided, put it on...
	if (hbmSkin)
	{

		HDC hdcSrc;
		HBITMAP hbmSrcOld;
		BITMAP bm;
		int innerWidth, innerHeight, bmInnerWidth, bmInnerHeight;

		// get info from the bitmap
		GetObject(hbmSkin, sizeof(BITMAP), &bm);

		// this is to prevent negative values on the width
		innerWidth = max(nWidth - (nBorderLeft + nBorderRight), 0);
		innerHeight = max(nHeight - (nBorderTop + nBorderBottom), 0);
		bmInnerWidth = max(bm.bmWidth - (nBorderLeft + nBorderRight), 1);
		bmInnerHeight = max(bm.bmHeight - (nBorderTop + nBorderBottom), 1);

		hdcSrc = CreateCompatibleDC(hdcDst);
		hbmSrcOld = (HBITMAP)SelectObject(hdcSrc, hbmSkin);

		if (nBorderTop || nBorderLeft || nBorderBottom || nBorderTop)
		{

			// paint the four corners first, and then the borders
			// topleft, topright, bottomright, bottomleft
			BltCopy(hdcDst, 0, 0, nBorderLeft, nBorderTop, hdcSrc, 0, 0);
			BltCopy(hdcDst, nWidth - nBorderRight, 0, nBorderRight, nBorderTop, hdcSrc, bm.bmWidth - nBorderRight, 0);
			BltCopy(hdcDst, nWidth - nBorderRight, nHeight - nBorderBottom, nBorderRight, nBorderBottom, hdcSrc, bm.bmWidth - nBorderRight, bm.bmHeight - nBorderBottom);
			BltCopy(hdcDst, 0, nHeight - nBorderBottom, nBorderLeft, nBorderBottom, hdcSrc, 0, bm.bmHeight - nBorderBottom);

			if (nBorderTop && innerWidth)
				BltSize(hdcDst, nBorderLeft, 0, innerWidth, nBorderTop, hdcSrc, nBorderLeft, 0, bmInnerWidth, nBorderTop);
			if (nBorderLeft && innerHeight)
				BltSize(hdcDst, 0, nBorderTop, nBorderLeft, innerHeight, hdcSrc, 0, nBorderTop, nBorderLeft, bmInnerHeight);
			if (nBorderRight && innerHeight)
				BltSize(hdcDst, nWidth - nBorderRight, nBorderTop, nBorderRight, innerHeight, hdcSrc, bm.bmWidth - nBorderRight, nBorderTop, nBorderRight, bmInnerHeight);
			if (nBorderBottom && innerWidth)
				BltSize(hdcDst, nBorderLeft, nHeight - nBorderBottom, innerWidth, nBorderBottom, hdcSrc, nBorderLeft, bm.bmHeight - nBorderBottom, bmInnerWidth, nBorderBottom);
		}

		// draw area in the middle
		if (innerWidth && innerHeight)
			BltSize(hdcDst, nBorderLeft, nBorderTop, innerWidth, innerHeight, hdcSrc, nBorderLeft, nBorderTop, bmInnerWidth, bmInnerHeight);

		SelectObject( hdcSrc, hbmSrcOld );
		DeleteDC( hdcSrc );
	}

	SelectObject( hdcDst, hbmDstOld );
	DeleteDC( hdcDst );

	if (!bTransSkin)
	{
		bTransBack = false;
	}
	else
	{
		HRGN hOpaqueRgn = CreateRectRgn(0, 0, nWidth, nHeight);
		hrgnBack = BitmapToRegion(hbmBack, RGB(255, 0, 255), 0x101010, 0, 0);

		// to see if any pixels are transparent
		bTransBack = !EqualRgn(hrgnBack, hOpaqueRgn);

		DeleteObject(hOpaqueRgn);

		if (!bTransBack)
		{
			DeleteObject(hrgnBack);
			hrgnBack = NULL;
		}
	}

	return bTransBack;
}

// paints the background and fills the HRGN with the bg region 
void BoxPainter::PaintBackground(HDC hdcDst, HRGN hrgnDst)
{

	HDC hdcSrc;
	HBITMAP hbmOld;

	// if the background isn't there, make one...
	if (!hbmBack)
	{
		CreateBackground();
	}
	else
	{
		BITMAP bm;
		GetObject(hbmBack, sizeof(BITMAP), &bm);
		if (bm.bmWidth != nWidth || bm.bmHeight != nHeight)
		{
			CreateBackground();
		}
	}

	// if there is no background now... do nothing
	if (!hbmBack)
		return ;


	hdcSrc = CreateCompatibleDC(hdcDst);
	hbmOld = (HBITMAP)SelectObject(hdcSrc, hbmBack);

	BitBlt(hdcDst, 0, 0, nWidth, nHeight, hdcSrc, 0, 0, SRCCOPY);

	SelectObject(hdcSrc, hbmOld);
	DeleteDC(hdcSrc);

	if ( hrgnBack && bTransBack )
		CombineRgn(hrgnDst, hrgnBack, NULL, RGN_COPY);
}

// grd Blitting functions...
// to enable trans / not trans painting
int BoxPainter::BltCopy(HDC hdcDst, int xDst, int yDst, int cxDst, int cyDst, HDC hdcSrc, int xSrc, int ySrc)
{
	if (bTransSkin)
		TransparentBltLS(hdcDst, xDst, yDst, cxDst, cyDst, hdcSrc, xSrc, ySrc, MAGICPINK);
	else
		return BitBlt(hdcDst, xDst, yDst, cxDst, cyDst, hdcSrc, xSrc, ySrc, SRCCOPY);

	return 0;
}

// to enable tile / stretch resizing
int BoxPainter::BltSize(HDC hdcDst, int xDst, int yDst, int cxDst, int cyDst, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc)
{
	HDC	hdcTmp;
	HBITMAP hbmTmp, hbmOld;
	int tmpX, tmpY;

	if (bTransSkin)
	{
		hdcTmp = CreateCompatibleDC( hdcDst );
		hbmTmp = CreateCompatibleBitmap( hdcDst, cxDst, cyDst );
		hbmOld = (HBITMAP)SelectObject( hdcTmp, hbmTmp );
		tmpX = 0, tmpY = 0;
	}
	else
	{
		hdcTmp = hdcDst;
		tmpX = xDst, tmpY = yDst;
	}

	if (bSkinTiled)
	{
		int x, y;
		// fill the first row of images
		for (x = 0; x < cxDst; x += cxSrc)
			BitBlt(hdcTmp, tmpX + x, tmpY, min(cxSrc, cxDst - x), min(cySrc, cyDst), hdcSrc, xSrc, ySrc, SRCCOPY);
		// copy this row down the y axis...
		for (y = cySrc; y < cyDst; y += cySrc)
			BitBlt(hdcTmp, tmpX, tmpY + y, cxDst, min(cySrc, cyDst - y), hdcTmp, tmpX, tmpY, SRCCOPY);

	}
	else
	{
		// stretch it
		StretchBlt(hdcTmp, tmpX, tmpY, cxDst, cyDst, hdcSrc, xSrc, ySrc, cxSrc, cySrc, SRCCOPY);
	}

	// blit the resulting image onto the destination dc
	if (bTransSkin)
	{
		TransparentBltLS(hdcDst, xDst, yDst, cxDst, cyDst, hdcTmp, tmpX, tmpY, MAGICPINK);

		// cleanup
		SelectObject( hdcTmp, hbmOld );
		DeleteObject( hbmTmp );
		DeleteDC( hdcTmp );
	}
	return 0;
}